local M = {}
local logTag = 'mathkuro'

-- 加速を行うダメージの閾値
local ACCEL_DAMAGE_THRESHOLD = 5000

-- 設定値(UI側との連携)
local settings = {
  acceleratePlayerVehicle = {
    enabled = false,
    speed = 0
  },
  accelerateNeabyTraffic = {
    enabled = false,
    speed = 0,
    radius = 0,
    excludeBrokenVehicles = false
  },
  forceField = {
    enabled = false,
    multiplier = 0
  },
  explodeNearbyVehicles = {
    enabled = false,
    radius = 0
  },
  dropObjects = {
    enabled = false,
    objectType = 'rocks',
    objectCount = 0,
    radius = 0,
    dropHeight = 0,
    intervalSec = 0
  },
  speedMode = {
    enabled = false,
    speed = 20,
    extensionTime = 10
  },
  crashCamera = {
    enabled = false
  }
}

-- クラッシュカメラの設定
local CrashCamera = {
  enabled = false,
  lastCameraName = 'orbit',
  lastPlayerDamage = -1,
  CRASH_CAMERA_MODE = 'external',
  CRASH_CAMERA_SIM_SPEED = 0.2,
  CRASH_CAMERA_DURATION_SEC = 8,
  CRASH_CAMERA_DAMAGE_THRESHOLD = 10000
}

function CrashCamera:enable()
  if self.enabled then
    if self.crashCameraFinishTime < os.clock() then
      -- クラッシュカメラ有効化から指定時間経過後は自動的に無効化
      self:disable()
    end
  elseif self:getPlayerDamageDelta() > self.CRASH_CAMERA_DAMAGE_THRESHOLD then
    -- 現在設定されているカメラを退避
    self.lastCameraName = core_camera.getActiveCamName()
    log('D', logTag, 'Crash Camera is enabled. lastCameraName:'..tostring(self.lastCameraName))

    bullettime.set(self.CRASH_CAMERA_SIM_SPEED)

    local _playerVelocity = be:getPlayerVehicle(0):getVelocity():length()
    log('D', logTag, '_playerVelocity:'.._playerVelocity)

    if _playerVelocity > 15 then
      core_camera.setByName(0, self.CRASH_CAMERA_MODE)
      core_camera.resetCamera(0)
    else
      core_camera.setByName(0, 'orbit')
      core_camera.cameraZoom(0.002)
      core_camera.rotate_yaw(0.1)
      local _vid = be:getPlayerVehicle(0):getId()
      -- core_camera.setFOV(_vid, 90)
    end

    self.enabled = true
    self.crashCameraFinishTime = os.clock() + self.CRASH_CAMERA_DURATION_SEC
  end
end

function CrashCamera:disable()
  if self.enabled then
    -- クラッシュカメラが有効化されている場合のみ解除処理を行う
    bullettime.set(1)
    core_camera.setByName(0, self.lastCameraName)
    core_camera.resetCamera(0)
    self.enabled = false
    log('D', logTag, 'Crash Camera is disabled')
  else
    log('D', logTag, 'crashCameraMode is not enabled')
  end
end

function CrashCamera:toggle(enabled)
  if enabled then
    self:enable()
  else
    self:disable()
  end
end

-- プレイヤーのダメージ差分(前回呼び出し時との差分)を取得
function CrashCamera:getPlayerDamageDelta()
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil or map.objects[_playerVeh:getId()] == nil then
    return 0
  end

  local _playerDamageDelta = 0
  local _playerDamage = map.objects[_playerVeh:getId()]["damage"]

  if _playerDamage < self.lastPlayerDamage then
    self.lastPlayerDamage = 0
  end

  if self.lastPlayerDamage == -1 then
    self.lastPlayerDamage = _playerDamage
    -- 初回呼び出し時に_playerDamageの値に関わらず差分は0として扱う
    _playerDamageDelta = 0
  else
    _playerDamageDelta = _playerDamage - self.lastPlayerDamage
    self.lastPlayerDamage = _playerDamage
  end

  return _playerDamageDelta
end

-- 車両ダメージを取得
local function getVehicleDamage(vid)
  local _currentDamage = 0
  local _veh = be:getObjectByID(vid)
  if map.objects[vid] == nil or map.objects[vid]['damage'] then
    -- 駐車されている車両はトラフィックのdamageが参照できないので独自にbeamstateから取得する
    _veh:queueLuaCommand('obj:queueGameEngineLua("be:getObjectByID(' .. vid .. ').destructionDamage = " .. beamstate.damage)')
    if _veh.destructionDamage == nil then
      _veh:setDynDataFieldbyName('destructionDamage', 0, 0)
      _currentDamage = 0
    else
      _currentDamage = _veh.destructionDamage
    end
  else
    _currentDamage = map.objects[vid]['damage']
  end

  return _currentDamage
end

local function acceleratePlayerVehicle(speed)
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil or map.objects[_playerVeh:getId()] == nil then
    return
  end

  _playerVeh:queueLuaCommand('thrusters.applyVelocity(obj:getDirectionVector() * ' .. tostring(speed) .. ')')
end

local function accelerateNeabyTraffic(speed, radiusMeter, excludeBrokenVehicles)
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil or map.objects[_playerVeh:getId()] == nil then
    return
  end

  local _playerPos = _playerVeh:getPosition()
  for vid, veh in activeVehiclesIterator() do
    -- 登録済みまたはプレイヤー車両との距離が閾値未満の車両のみダメージ値更新
    if vid ~= _playerVeh:getId() and (veh:getPosition():distance(_playerPos) < radiusMeter) then
      -- ダメージが閾値以下の車両(ダメージ未登録車両を含む)を加速
      if not excludeBrokenVehicles or getVehicleDamage(vid) < ACCEL_DAMAGE_THRESHOLD then
        veh:queueLuaCommand('thrusters.applyVelocity(obj:getDirectionVector() * ' .. tostring(speed) .. ')')
      end
    end
  end
end

local function adjustForceField(enabled, multiplier)
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil or map.objects[_playerVeh:getId()] == nil then
    return
  end

  if enabled then
    extensions.gameplay_forceField.activate()
    extensions.gameplay_forceField.setForceMultiplier(multiplier)
  else
    extensions.gameplay_forceField.deactivate()
  end
end

local function explodeNearbyVehicles(radiusMeter)
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil or map.objects[_playerVeh:getId()] == nil then
    return
  end

  local _playerPos = _playerVeh:getPosition()
  for vid, veh in activeVehiclesIterator() do
    -- 登録済みまたはプレイヤー車両との距離が閾値未満の車両のみダメージ値更新
    if vid ~= _playerVeh:getId() and (veh:getPosition():distance(_playerPos) < radiusMeter) then
      veh:queueLuaCommand('fire.explodeVehicle()')
      veh:queueLuaCommand('beamstate.breakAllBreakgroups()')
    end
  end
end

local DropObjectGroup = {}
DropObjectGroup.groupId = nil
DropObjectGroup.vids = {}

DropObjectGroup.objectCount = 0
DropObjectGroup.objectType = 'rocks'
DropObjectGroup.radiusMeter = 30
DropObjectGroup.dropHeight = 30
DropObjectGroup.intervalSec = 10
DropObjectGroup.respawnAlways = true
DropObjectGroup.useForwardDrop = true
DropObjectGroup.HEIGHT_COEF = 1
DropObjectGroup.STASH_FILE = '/temp/mathkuro/chaosModInfo.json'

function DropObjectGroup:init()
  local _vids = jsonReadFile(self.STASH_FILE)
  if _vids == nil or _vids == {} then
    self.vids = {}
    self.objectType = 'rocks'
    self.objectCount = 0
    log('D', logTag, 'DropObjectGroup:init() _vids is nil')
  else
    self.vids = {}
    self.objectType = table.remove(_vids)
    for _, vid in ipairs(_vids) do
      if be:getObjectByID(vid) then
        table.insert(self.vids, vid)
      end
    end
    self.objectCount = #self.vids
  end
end

function DropObjectGroup:writeFile(vids, objectType)
  local _vids = deepcopy(vids)
  table.insert(_vids, objectType)
  jsonWriteFile(self.STASH_FILE, _vids, true)
end

function DropObjectGroup:dropObjects(simTime)
  local _playerVeh = be:getPlayerVehicle(0)
  if _playerVeh == nil or map.objects[_playerVeh:getId()] == nil then
    return
  end

  local _playerPos = _playerVeh:getPosition()
  local _playerVelocity = _playerVeh:getVelocity()
  local _coef = 0
  if self.useForwardDrop then
    -- 前気味に落としたい場合に使用する係数。落下高度(地面に到達する時刻) (s = 1/2 * g * t^2 -> t = sqrt(2 * s / g))
    _coef = math.sqrt(2 * self.dropHeight / math.abs(core_environment.getGravity())) * 1.2
  end

  for _, vid in ipairs(self.vids) do
    local veh = be:getObjectByID(vid)
    if veh and (veh.nextSpawnSec == nil or veh.nextSpawnSec < simTime) and (self.respawnAlways or (veh:getPosition():distance(_playerPos) > self.radiusMeter / 2)) then

      -- プレイヤーの位置+速度(プレイヤーの進行方向)を基準に、到達予想時刻度による係数をかけた位置を中心にスポーン地点を設定する
      local _pos = _playerPos + _playerVelocity * _coef
      _pos.x = _pos.x + (math.random() - 0.5) * 2 * self.radiusMeter
      _pos.y = _pos.y + (math.random() - 0.5) * 2 * self.radiusMeter
      _pos.z = _pos.z + self.dropHeight
      local _rot = quat(math.random(), math.random(), math.random(), math.random())
      spawn.safeTeleport(veh, _pos, _rot)

      -- 次スポーン時刻を設定(インターバル経過後の前後2秒)
      veh.nextSpawnSec = simTime + (self.intervalSec - 2) + 4 * math.random()
    end
  end
end

function DropObjectGroup:deleteAll()
  if self.vids ~= {} then
    for _, vid in ipairs(self.vids) do
      local veh = be:getObjectByID(vid)
      if veh then
        veh:delete()
      end
    end
    self.vids = {}
    self.objectCount = 0
    self:writeFile(self.vids, self.objectType)
  end
end

local function toggleDropObjects(enabled, objectType, objectCount, radiusMeter, dropHeight, intervalSec, respawnAlways, useForwardDrop)
  log('D', logTag, 'toggleDropObjects('..tostring(enabled)..', '..objectType..', '..objectCount..', '..radiusMeter..', '..dropHeight..', '..intervalSec..', '..tostring(respawnAlways)..', '..tostring(useForwardDrop)..')')
  if enabled then
    if objectType ~= DropObjectGroup.objectType then
      DropObjectGroup:deleteAll()
    end

    local spawnCount = objectCount - DropObjectGroup.objectCount
    if spawnCount > 0 then
      local _group = {}

      if objectType == "randomVehicle" then
        _group = extensions.core_multispawn.createGroup(spawnCount)
      elseif objectType == "randomSimpleVehicle" then
        _group = extensions.core_multispawn.createGroup(spawnCount, {filters = {Type = {propparked = 1}}, allConfigs = true})
      elseif objectType == "randomProp" then
        _group = extensions.core_multispawn.createGroup(spawnCount, {filters = {Type = {prop = 1}}, allConfigs = true})
      else
        -- デフォルトは'rocks'
        _group = extensions.core_multispawn.createGroup(spawnCount, {filters = {key = {rock1 = 1, rock2 = 1, rock3 = 1, rock4 = 1, rock5 = 1, rock6 = 1, rock7 = 1}}, allConfigs = true})
      end

      -- 一時停止してからスポーン
      simTimeAuthority.pause(true)
      DropObjectGroup.groupId = extensions.core_multispawn.spawnGroup(_group, spawnCount)
    elseif spawnCount < 0 then
      local deleteCount = 0
      local remainVids = {}
      for _, vid in ipairs(DropObjectGroup.vids) do
        local veh = be:getObjectByID(vid)
        if veh then
          if deleteCount < -spawnCount then
            veh:delete()
            deleteCount = deleteCount + 1
          else
            table.insert(remainVids, vid)
          end
        end
      end
      DropObjectGroup.vids = remainVids
      DropObjectGroup:writeFile(DropObjectGroup.vids, DropObjectGroup.objectType)
    end

    DropObjectGroup.objectCount = objectCount
    DropObjectGroup.objectType = objectType
    DropObjectGroup.radiusMeter = radiusMeter
    DropObjectGroup.dropHeight = dropHeight
    DropObjectGroup.intervalSec = intervalSec
    DropObjectGroup.respawnAlways = respawnAlways
    DropObjectGroup.useForwardDrop = useForwardDrop
  else
    DropObjectGroup:deleteAll()
  end
end

local SpeedMode = {}
SpeedMode.enabled = false
SpeedMode.minimumSpeedLimit = 20
SpeedMode.extensionTime = 10
SpeedMode.remainingTime = SpeedMode.extensionTime
SpeedMode.lastDisplayTime = SpeedMode.extensionTime
SpeedMode.deadLineTime = 0
function SpeedMode:toggle(enabled, minimumSpeedLimit, extensionTime)
  self.enabled = enabled
  self.minimumSpeedLimit = minimumSpeedLimit
  self.extensionTime = extensionTime
  self:reset()
end

function SpeedMode:reset()
  self.remainingTime = self.extensionTime
  self.lastDisplayTime = self.extensionTime
  self.deadLineTime = 0
  guihooks.trigger('ScenarioRealtimeDisplay', {})
end

function SpeedMode:update(totalSimTime)
  local _playerVelocity = be:getPlayerVehicle(0):getVelocity():length()
  if _playerVelocity >= self.minimumSpeedLimit then
    self:reset()
    return
  end

  if self.deadLineTime == 0 then
    self.deadLineTime = totalSimTime + self.extensionTime
  end

  local remainingTime = self.deadLineTime - totalSimTime
  log('D', logTag, 'SpeedMode:update('..totalSimTime..') remainingTime:'..remainingTime)
  if self.lastDisplayTime - 1 < remainingTime then
    return
  end

  self.lastDisplayTime = math.ceil(remainingTime)

  local displayVelocity, unit = translateVelocity(self.minimumSpeedLimit, true)

  local msgStr = 'MUST ACCELERATE UP TO ' .. tostring(math.ceil(displayVelocity)) .. unit .. '! ' ..tostring(self.lastDisplayTime)..' SECONDS LEFT!'

  guihooks.trigger('Message', {ttl = 1, msg = msgStr, category = 'mathkuroChaosModUI'})
  guihooks.trigger('ScenarioRealtimeDisplay', {msg = msgStr})
  Engine.Audio.playOnce('AudioGui', 'event:UI_Countdown1')

  if remainingTime <= 0 then
    Engine.Audio.playOnce('AudioGui', 'event:UI_Checkpoint')
    be:getPlayerVehicle(0):queueLuaCommand("fire.explodeVehicle()")
    be:getPlayerVehicle(0):queueLuaCommand("beamstate.breakAllBreakgroups()")
    self:reset()
  end
end


local function applySettings(_settings)
  log('D', logTag, 'applySettings('..dumps(_settings)..')')
  settings = deepcopy(_settings)

  adjustForceField(settings.forceField.enabled, settings.forceField.multiplier)

  toggleDropObjects(
    settings.dropObjects.enabled, settings.dropObjects.objectType,
    settings.dropObjects.objectCount, settings.dropObjects.radius,
    settings.dropObjects.dropHeight, settings.dropObjects.intervalSec,
    settings.dropObjects.respawnAlways, settings.dropObjects.useForwardDrop)

  SpeedMode:toggle(settings.speedMode.enabled, settings.speedMode.speed, settings.speedMode.extensionTime)

  CrashCamera:toggle(settings.crashCamera.enabled)
end

local function onInit()
  log('D', logTag, 'onInit()')

  DropObjectGroup:init()
end

local function onVehicleGroupSpawned(vids, groupId)
  log('D', logTag, 'groupId:'..groupId..', DropObjectGroup.groupId:'..tostring(DropObjectGroup.groupId))
  if DropObjectGroup.groupId == groupId then
    for _, vid in ipairs(vids) do
      table.insert(DropObjectGroup.vids, vid)
    end
    log('D', logTag, 'DropObjectGroup.vids:'..dumps(DropObjectGroup.vids))
    DropObjectGroup.groupId = nil

    DropObjectGroup:writeFile(DropObjectGroup.vids, DropObjectGroup.objectType)
  end
end

local execThreshold = 1 / 10
local tempSimTime = 0
local totalSimTime = 0
local function onUpdate(dtReal, dtSim, dtRaw)
  tempSimTime = tempSimTime + dtSim
  totalSimTime = totalSimTime + dtSim
  if tempSimTime < execThreshold then
    -- 一定間隔でのみ実行&ポーズ中は実行しない(ポーズ中はdtSimは進まない)
    return
  end
  -- log('D', logTag, 'onUpdate('..dtReal..', '..dtSim..', '..dtRaw..') executed. tempSimTime:'..tempSimTime..', totalSimTime:'..totalSimTime)
  tempSimTime = 0

  if settings.acceleratePlayerVehicle.enabled then
    acceleratePlayerVehicle(settings.acceleratePlayerVehicle.speed)
  end

  if settings.accelerateNeabyTraffic.enabled then
    accelerateNeabyTraffic(settings.accelerateNeabyTraffic.speed, settings.accelerateNeabyTraffic.radius, settings.accelerateNeabyTraffic.excludeBrokenVehicles)
  end

  if settings.explodeNearbyVehicles.enabled then
    explodeNearbyVehicles(settings.explodeNearbyVehicles.radius)
  end

  if settings.dropObjects.enabled then
    DropObjectGroup:dropObjects(totalSimTime)
  end

  if settings.speedMode.enabled then
    SpeedMode:update(totalSimTime)
  end

  if settings.crashCamera.enabled then
    CrashCamera:enable()
  end
end

local function unload()
  log('D', logTag, 'unload()')
  extensions.unload(M)
end


M.applySettings = applySettings

M.onInit = onInit
M.onVehicleGroupSpawned = onVehicleGroupSpawned
M.onUpdate = onUpdate
M.unload = unload

return M
