
angular
.module('beamng.apps')
.controller('AppCtrl', function ($scope, $mdSidenav) {
  // ---- 定数 ----
  const SETTINGS_PATH = '/ui/modules/apps/chaosModMathkuro/settings.json';

  // 言語設定
  const DEFAULT_LANG = 'en';
  const SUPPORT_LANGS = ['en', 'ja'];
  let currentLanguage = DEFAULT_LANG;

  // ---- UI変数 ----
  // UIの表示設定・個別にON/OFF、ファイル保存
  $scope.settings = {
    acceleratePlayerVehicle: {
      enabled: false,
      speed: 50
    },
    accelerateNeabyTraffic: {
      enabled: false,
      speed: 50,
      radius: 20,
      excludeBrokenVehicles: true
    },
    forceField: {
      enabled: false,
      multiplier: 1.0
    },
    explodeNearbyVehicles: {
      enabled: false,
      radius: 20
    },
    dropObjects: {
      enabled: false,
      objectCount: 50,
      objectType: 'rocks',
      radius: 25,
      dropHeight: 25,
      intervalSec: 8,
      respawnAlways: true,
      useForwardDrop: true
    },
    speedMode: {
      enabled: false,
      speed: 20,
      extensionTime: 10
    },
    crashCamera: {
      enabled: false
    }
  }
  // ドロップオブジェクトの種別
  $scope.dropObjectsTypes = [
    { name: 'Rock', value: 'rocks' },
    { name: 'Prop', value: 'randomProp' },
    { name: 'Vehicle', value: 'randomVehicle' },
    { name: 'Simple Vehicle', value: 'randomSimpleVehicle' }
  ]

  // luaのロード
  bngApi.engineLua("extensions.load('mathkuro_chaosMod')");

  // 設定ファイル読み込み
  bngApi.engineLua('jsonReadFile(' + bngApi.serializeToLua(SETTINGS_PATH) + ')', (settings) => {
    if (settings != null) {
      console.log(settings);
      $scope.settings = settings;
    }

    // Lua側に設定を反映
    bngApi.engineLua('mathkuro_chaosMod.applySettings(' + bngApi.serializeToLua($scope.settings) + ')');
  })

  // settings/Closeボタン押下時に実行される関数
  $scope.toggleLeft = buildToggler('left');
  $scope.closeLeft = closeToggler('left');
  // サイドメニューの開閉
  function buildToggler(componentId) {
    return function() {
      $mdSidenav(componentId).toggle();
    }
  }
  function closeToggler(componentId) {
    return function() {
      $mdSidenav(componentId).toggle();
      // サイドメニューを閉じた際に設定を保存しておく
      bngApi.engineLua('jsonWriteFile(' + bngApi.serializeToLua(SETTINGS_PATH) + ', ' + bngApi.serializeToLua($scope.settings) + ', ' + bngApi.serializeToLua(true) + ')');

      // Lua側に設定を反映
      bngApi.engineLua('mathkuro_chaosMod.applySettings(' + bngApi.serializeToLua($scope.settings) + ')');
    }
  }

  $scope.unit = function(...args) { return UiUnits.buildString(...args); }

  // Get user language settings. ('<lang>_<country>')
  // ref: 0.25\lua\common\utils\languageMap.lua
  bngApi.engineLua('Lua:getSelectedLanguage()', (userLang) => {
    let _lang = userLang.split('_')[0];
    currentLanguage = (SUPPORT_LANGS.includes(_lang)) ? _lang : DEFAULT_LANG;
    console.log('set language: ' + currentLanguage);

    // apps上の翻訳表示
    $scope.ui = getLocale(currentLanguage).UI;
    $scope.ui.speedUnit =  UiUnits.speed(0).unit;
    $scope.ui.lengthUnit =  UiUnits.length(10).unit;
  });

  // ---- UI上に表示する言語の設定 ----
  function getLocale(localeName){
    if (localeName == 'ja') {
      return new JaLocale();
    } else {
      return new EnLocale();
    }
  };

  class LocalInfoBase {
    UI = {
      closeButtonText: 'Apply and Close',

      chaosModSettingsTitle: 'Chaos Mod Settings',
      acceleratePlayerVehicle: {
        title: 'Accelerate Player Vehicle',
        speed: 'Speed'
      },
      accelerateNeabyTraffic: {
        title: 'Accelerate Nearby Traffic',
        speed: 'Speed',
        radius: 'Radius',
        excludeBrokenVehicles: 'Exclude Broken Vehicles'
      },
      forceField: {
        title: 'Force Field',
        multiplier: 'Multiplier'
      },
      explodeNearbyVehicles: {
        title: 'Explode Nearby Vehicles',
        radius: 'Radius'
      },
      dropObjects: {
        title: 'Drop Objects',
        objectCount: 'Object Count',
        objectType: 'Object Type',
        radius: 'Radius',
        dropHeight: 'Drop Height',
        intervalSec: 'Interval Sec',
        respawnAlways: 'Respawn Always',
        respawnAlwaysTooltip: 'When not set, objects close to the player will not despawn',
        useForwardDrop: 'Use Forward Drop'
      },
      speedMode: {
        title: 'Speed Mode',
        speed: 'Speed',
        extensionTime: 'Extension Time'
      },

      UIsettingsTitle: 'UI Settings',
      crashCamera: {
        title: 'Crash Camera'
      }
    }
  };

  class EnLocale extends LocalInfoBase {};

  class JaLocale extends LocalInfoBase {
    UI = {
      closeButtonText: '適用して閉じる',

      chaosModSettingsTitle: 'カオスMOD設定',
      acceleratePlayerVehicle: {
        title: '強制加速(プレイヤー)',
        speed: '速度'
      },
      accelerateNeabyTraffic: {
        title: '強制加速(周辺車両)',
        speed: '速度',
        radius: '影響範囲',
        excludeBrokenVehicles: '故障車両を除外'
      },
      forceField: {
        title: 'フォースフィールド',
        multiplier: '倍率'
      },
      explodeNearbyVehicles: {
        title: '近くの車両爆破',
        radius: '影響範囲'
      },
      dropObjects: {
        title: 'オブジェクト無限落下',
        objectCount: 'オブジェクト数',
        objectType: 'オブジェクト種別',
        radius: '影響範囲',
        dropHeight: '落下高度',
        intervalSec: '間隔',
        respawnAlways: 'オブジェクトを常にリスポーン',
        respawnAlwaysTooltip: '未設定時はプレイヤーに近接しているオブジェクトはデスポーンしません',
        useForwardDrop: '前方への落下'
      },
      speedMode: {
        title: 'Speedモード',
        speed: '速度',
        extensionTime: '残り時間'
      },
      UIsettingsTitle: 'UI設定',
      crashCamera: {
        title: 'クラッシュカメラ'
      }
    }
  };

})
.directive('chaosModMathkuro', [function () {
  return {
    templateUrl: '/ui/modules/apps/chaosModMathkuro/app.html',
    replace: true,
    restrict: 'EA',
    link: function (scope, element, attrs) {
      'use strict';
    }
  }
}])
